////////////////////////////////////////////////////////////////////////
// Class:       OnlineWaveform
// Module Type: producer
// File:        OnlineWaveform_module.cc
//
// Generated at Sat Jun 29 21:28:28 2013 by Alessandro Razeto using artmod
// from cetpkgsupport v1_02_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "artdaq-core/Data/Fragments.hh"
#include "darksidecore/Data/V172xFragment.hh"
#include "darkmon/ArtModules/OnlineEvent.hh"
#include "darksidecore/ArtServices/DBInterface.hh"

#include <memory>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <array>

namespace ds50 {
  class OnlineWaveform : public art::EDProducer {
    public:
      explicit OnlineWaveform(fhicl::ParameterSet const & p);
      virtual ~OnlineWaveform() {}

      void produce(art::Event & e) override;
      void beginRun(art::Run &) override;

    private:
      std::array<float, 40> gains_;
      std::string board_type_;
      std::vector<float> baseline_limits_us_;
  };
}


ds50::OnlineWaveform::OnlineWaveform(fhicl::ParameterSet const & ps): board_type_(ps.get<std::string> ("board_type", "V1720")), baseline_limits_us_(ps.get<std::vector<float>> ("baseline_limits_us", std::vector<float>(0))) {
  produces<ds50::Online::Waveforms> ();
}

void ds50::OnlineWaveform::produce(art::Event &e) {
  art::Handle<artdaq::Fragments> v172x;
  e.getByLabel ("daq", board_type_, v172x);
  if (!v172x.isValid ())  {
    e.getByLabel ("unhuff" + board_type_, board_type_, v172x);
    if (!v172x.isValid ()) return;
  }

  std::unique_ptr<Online::Waveforms> wf(new Online::Waveforms);

  size_t record_size = 100000000;
  for (size_t i = 0; i < v172x->size(); ++i) record_size = std::min (record_size, V172xFragment((*v172x)[i]).adc_values_for_channel());

  for (size_t i = 0; i < v172x->size(); ++i) {
    const auto& frag((*v172x)[i]);

    if (!frag.hasMetadata ()) return;
    const V172xFragment::metadata *md = frag.metadata<V172xFragment::metadata> ();

    if (!i) {
      wf->sample_ns = 1000 / md->sample_rate_MHz;
      wf->post_samples = record_size * md->post_trigger / 100;
      wf->pre_samples = record_size - wf->post_samples;
      wf->sum.resize (record_size);
      wf->times_ns.resize (record_size);
      float offset = -wf->pre_samples * wf->sample_ns;
      for (float &x: wf->times_ns) { x = offset; offset += wf->sample_ns; }
      if (baseline_limits_us_.size () && (baseline_limits_us_[0] >= baseline_limits_us_[1] || baseline_limits_us_[0] < wf->times_ns.front () || baseline_limits_us_[1] > wf->times_ns.back ()))
	throw cet::exception("OnlineWaveform") << "baseline_limits_us out of range (" << baseline_limits_us_[0] << "," << baseline_limits_us_[1] << ")";
    }

    int limits[2] = { 0, wf->pre_samples - int(md->sample_rate_MHz * 0.5) };
    if (baseline_limits_us_.size ()) {
      limits[0] = baseline_limits_us_[0] * md->sample_rate_MHz + wf->pre_samples;
      limits[1] = baseline_limits_us_[1] * md->sample_rate_MHz + wf->pre_samples;
      if (limits[0] < 0 || limits[1] < 0) throw cet::exception("OnlineCluster") << "baseline_limits_us outside the gate";
    }

    V172xFragment b(frag);
    size_t board_id = b.board_id ();

    for (int ch = 0; ch < 8; ++ch) {
      int lg = board_id * 8 + ch;
      if (!b.channel_present (ch) || lg >= 40 || gains_[lg] <= 0) continue;
      wf->wf[lg].resize (record_size);
      std::copy(b.chDataBegin (ch), b.chDataBegin (ch) + record_size, wf->wf[lg].begin());
      std::for_each (wf->wf[lg].begin(), wf->wf[lg].end(), [&](float &v) { v /= -gains_[lg]; }); 
      float baseline = std::accumulate (wf->wf[lg].begin() + limits[0], wf->wf[lg].begin() + limits[1], 0.) / (limits[1] - limits[0]);
      for (float &x: wf->wf[lg]) x -= baseline;
      std::transform (wf->sum.begin (), wf->sum.end (), wf->wf[lg].begin (), wf->sum.begin (), std::plus<float>());
    }
  }
  e.put (std::move(wf));

}

void ds50::OnlineWaveform::beginRun(art::Run &) {
  art::ServiceHandle<ds50::DBInterface> dbi;
  ds50::db::result r = dbi->latest ("shadowside.ser", "");

  if (r.cell_elements("spe_mean", 0) != gains_.size ())
    throw cet::exception("OnlineWaveform") << "wrong number of elements";

  for (size_t i = 0; i < r.cell_elements("spe_mean", 0); i++) gains_[i] = r.get<float> ("spe_mean", 0, i);
}


DEFINE_ART_MODULE(ds50::OnlineWaveform)
